@@ -22,26 +22,43 @@ class GMusicLibraryProvider(base.BaseLibraryProvider):
                 values = [values]
             # FIXME this is bound to be slow for large libraries
             for value in values:
-                q = value.strip()
+                if field == 'track_no':
+                    q = self._convert_to_int(value)
+                else:
+                    q = value.strip()
 
                 uri_filter = lambda t: q == t.uri
-                track_filter = lambda t: q == t.name
+                track_name_filter = lambda t: q == t.name
                 album_filter = lambda t: q == getattr(t, 'album', Album()).name
                 artist_filter = lambda t: filter(
-                    lambda a: q == a.name, t.artists)
+                    lambda a: q == a.name, t.artists) or filter(
+                    lambda a: q == a.name, getattr(t, 'album',
+                                                   Album()).artists)
+                albumartist_filter = lambda t: any([
+                    q == a.name
+                    for a in getattr(t.album, 'artists', [])])
+                track_no_filter = lambda t: q == t.track_no
                 date_filter = lambda t: q == t.date
                 any_filter = lambda t: (
-                    track_filter(t) or album_filter(t) or
-                    artist_filter(t) or uri_filter(t))
+                    uri_filter(t) or
+                    track_name_filter(t) or
+                    album_filter(t) or
+                    artist_filter(t) or
+                    albumartist_filter(t) or
+                    date_filter(t))
 
                 if field == 'uri':
                     result_tracks = filter(uri_filter, result_tracks)
-                elif field == 'track':
-                    result_tracks = filter(track_filter, result_tracks)
+                elif field == 'track_name':
+                    result_tracks = filter(track_name_filter, result_tracks)
                 elif field == 'album':
                     result_tracks = filter(album_filter, result_tracks)
                 elif field == 'artist':
                     result_tracks = filter(artist_filter, result_tracks)
+                elif field == 'albumartist':
+                    result_tracks = filter(albumartist_filter, result_tracks)
+                elif field == 'track_no':
+                    result_tracks = filter(track_no_filter, result_tracks)
                 elif field == 'date':
                     result_tracks = filter(date_filter, result_tracks)
                 elif field == 'any':
@@ -62,6 +79,11 @@ class GMusicLibraryProvider(base.BaseLibraryProvider):
             return []
 
     def _lookup_track(self, uri):
+        if uri.startswith('gmusic:track:T'):
+            song = self.backend.session.get_track_info(uri.split(':')[2])
+            if song is None:
+                return []
+            return [self._aa_to_mopidy_track(song)]
         try:
             return [self.tracks[uri]]
         except KeyError:
@@ -112,26 +134,45 @@ class GMusicLibraryProvider(base.BaseLibraryProvider):
                 values = [values]
             # FIXME this is bound to be slow for large libraries
             for value in values:
-                q = value.strip().lower()
+                if field == 'track_no':
+                    q = self._convert_to_int(value)
+                else:
+                    q = value.strip().lower()
 
                 uri_filter = lambda t: q in t.uri.lower()
-                track_filter = lambda t: q in t.name.lower()
+                track_name_filter = lambda t: q in t.name.lower()
                 album_filter = lambda t: q in getattr(
                     t, 'album', Album()).name.lower()
                 artist_filter = lambda t: filter(
-                    lambda a: q in a.name.lower(), t.artists)
+                    lambda a: q in a.name.lower(), t.artists) or filter(
+                    lambda a: q in a.name, getattr(t, 'album',
+                                                   Album()).artists)
+
+                albumartist_filter = lambda t: any([
+                    q in a.name.lower()
+                    for a in getattr(t.album, 'artists', [])])
+                track_no_filter = lambda t: q == t.track_no
                 date_filter = lambda t: t.date and t.date.startswith(q)
-                any_filter = lambda t: track_filter(t) or album_filter(t) or \
-                    artist_filter(t) or uri_filter(t)
+                any_filter = lambda t: (
+                    uri_filter(t) or
+                    track_name_filter(t) or
+                    album_filter(t) or
+                    artist_filter(t) or
+                    albumartist_filter(t) or
+                    date_filter(t))
 
                 if field == 'uri':
                     result_tracks = filter(uri_filter, result_tracks)
-                elif field == 'track':
-                    result_tracks = filter(track_filter, result_tracks)
+                elif field == 'track_name':
+                    result_tracks = filter(track_name_filter, result_tracks)
                 elif field == 'album':
                     result_tracks = filter(album_filter, result_tracks)
                 elif field == 'artist':
                     result_tracks = filter(artist_filter, result_tracks)
+                elif field == 'albumartist':
+                    result_tracks = filter(albumartist_filter, result_tracks)
+                elif field == 'track_no':
+                    result_tracks = filter(track_no_filter, result_tracks)
                 elif field == 'date':
                     result_tracks = filter(date_filter, result_tracks)
                 elif field == 'any':
@@ -208,5 +249,51 @@ class GMusicLibraryProvider(base.BaseLibraryProvider):
         self.artists[uri] = artist
         return artist
 
+    def _aa_to_mopidy_track(self, song):
+        uri = 'gmusic:track:' + song['storeId']
+        album = self._aa_to_mopidy_album(song)
+        artist = self._aa_to_mopidy_artist(song)
+        return Track(
+            uri=uri,
+            name=song['title'],
+            artists=[artist],
+            album=album,
+            track_no=song.get('trackNumber', 1),
+            disc_no=song.get('discNumber', 1),
+            date=album.date,
+            length=int(song['durationMillis']),
+            bitrate=320)
+
+    def _aa_to_mopidy_album(self, song):
+        album_info = self.backend.session.get_album_info(song['albumId'],
+                                                         include_tracks=False)
+        if album_info is None:
+            return None
+        name = album_info['name']
+        artist = self._aa_to_mopidy_album_artist(album_info)
+        date = unicode(album_info.get('year', 0))
+        return Album(
+            name=name,
+            artists=[artist],
+            date=date)
+
+    def _aa_to_mopidy_artist(self, song):
+        name = song['artist']
+        return Artist(
+            name=name)
+
+    def _aa_to_mopidy_album_artist(self, album_info):
+        name = album_info.get('albumArtist', '')
+        if name.strip() == '':
+            name = album_info['artist']
+        return Artist(
+            name=name)
+
     def _create_id(self, u):
         return hashlib.md5(u.encode('utf-8')).hexdigest()
+
+    def _convert_to_int(self, string):
+        try:
+            return int(string)
+        except ValueError:
+            return object()
